#
# Makefile to build the protected mode portion of the graphics
# demo program with the glue routines.  We use the High C-386
# compiler driver which comes with versions 1.6 and later of
# the High C compiler to compile the C code and link the whole
# image together with the compiler runtime libraries.
#

#
# Macro for the compiler name.  If the compiler is in auto mode
# set this to hc386.  If the compiler is in manual mode, set this
# to hcdriver.
#
hc = hc386	# works when the compiler is in auto mode
#hc = hcdriver	# works when the compiler is in manual mode

#
# Build the main demonstration program.  This program does not change,
# regardless of the strategy being used for intermode communications.
#
gdemo.obj: gdemo.c
	$(hc) $*.c -c -Hpro=..\..\includes\pl386.pro -Hoff=Floating_point

#
# Build glue routines for both REALBREAK and CALLBUFS
# intermode call methods.
#
gglue_rb.obj: gglue.c
	$(hc) gglue.c -c -Hpro=..\..\includes\pl386.pro -DREALBREAK -Hobject=gglue_rb.obj -Hoff=Floating_point

gglue_cb.obj: gglue.c
	$(hc) gglue.c -c -Hpro=..\..\includes\pl386.pro -Hobject=gglue_cb.obj -Hoff=Floating_point

prot_rb.obj: prot.asm
	386asm prot -nolist -d REALBREAK -o prot_rb.obj

prot_cb.obj: prot.asm
	386asm prot -nolist -o prot_cb.obj

#
# Build versions of the demo which use both the REALBREAK and CALLBUFS
# intermode call methods.
#
gdemo_rb.exp: gdemo.obj gglue_rb.obj prot_rb.obj
	$(hc) prot_rb.obj gdemo.obj gglue_rb.obj -o gdemo_rb.exp -Hl=e "-realbreak _realend" -sym -nomap "-maxreal 2000h"

gdemo_cb.exp: gdemo.obj gglue_cb.obj prot_cb.obj
	$(hc) gdemo.obj gglue_cb.obj prot_cb.obj -o gdemo_cb.exp -Hl=e "-callbufs 4" -sym -nomap "-maxreal 2000h"
