	.386r
;
;	real.asm - Real mode assembler code for gserver
;
;************************************************************************/
;*	Copyright (C) 1986-1988 Phar Lap Software, Inc.			*/
;*	Unpublished - rights reserved under the Copyright Laws of the	*/
;*	United States.  Use, duplication, or disclosure by the 		*/
;*	Government is subject to restrictions as set forth in 		*/
;*	subparagraph (c)(1)(ii) of the Rights in Technical Data and 	*/
;*	Computer Software clause at 252.227-7013.			*/
;*	Phar Lap Software, Inc., 60 Aberdeen Ave., Cambridge, MA 02138	*/
;************************************************************************/


	extrn	_gserver:far,_exit:far

	assume	cs:codeseg

codeseg	segment public byte 'code'


;
;	Variables used by this assembly language code
;

dssave	dw	?		; Real mode DS register
sssave	dw	?		; Real mode SS register
essave	dw	?		; Real mode ES register
spsave	dw	?		; Real mode SP register
bpsave	dw	?		; Real mode BP register

realpsp	dw	?		; Real mode PSP pointer
protpsp	dw	?		; Protected mode PSP pointer

spsave2 dw	?		; SP register provided by DOS-Extender
sssave2	dw	?		; SS register provided by DOS-Extender

reptr	dd	_reenter	; Far pointer to _reenter

;
;	protret - Return from real mode back to protected mode
;
;	This routine is called by the real mode server when
;	initialization of the server has been completed.  It returns
;	to the protected mode code which loaded the real mode server.
;	It returns by calling the 386|DOS-Extender protected mode
;	call function.  The address of this function as well as the
;	protected mode return address are taken from the parameter
;	block that is passed into the real mode server as the command
;	tail by the protected mode code.
;

	public	_protret

_protret proc	far

;	First, the real mode PSP segment pointer, the real mode segment
;	registers, and the real mode stack are saved for future calls
;	to the real mode server.

	mov	ah,62h		; Get the address of the PSP into ES.
	int	21h		;
	mov	es,bx		;

	mov	realpsp,bx	; Save the real mode PSP pointer.

	mov	dssave,ds	; Save the current real mode registers.
	mov	sssave,ss	;
	mov	essave,es	;
	mov	bpsave,bp	;
	mov	ax,sp		;
	sub	ax,20h		;
	mov	spsave,ax	;

; 	The following sanity check is done to make sure that the
;	we were called by a protected mode program

	mov	al,es:80h	; Branch if the length of the command
	cmp	al,12		;    tail is not 12.
	jne	#err		;

;	Save the segment address of the protected mode PSP for future
;	calls into the real mode code.

	mov	ax,es:8bh	; Save the protected mode PSP.
	mov	protpsp,ax	;

;	Return to protected mode via 386|DOS-Extender

	sub	ax,ax		; Zero the parameter block pointer on
	push	ax		;    the stack.
	push	ax		;

	mov	ax,es:89h	; Push the prot mode CS onto
	push	ax		;    the stack.

	mov	ax,es:87h	; Push the protected mode return address
	push	ax		;    onto the stack.
	mov	ax,es:85h	;
	push	ax		;

	mov	eax,reptr	; Load a far pointer to the re-enter routine
				;    into EAX.

	call	es:(dword ptr 81h) ; Call back into protected mode.

;	When the protected mode code returns to this point, it wants us
;	to exit.

;	Switch to the real mode PSP so that MS-DOS is all happy.

	mov	ah,50h		; Restore the real mode PSP.
	mov	bx,realpsp	;
	int	21h		;

;	Call the C exit routine to terminate and go back to protected mode
;	forever.

	push	0		; Call the C exit routine.
	call	_exit		;


;	Fatal error handler.

#err:	mov	dx,offset #errmsg; Output a fatal error message.
	push	cs		;
	pop	ds		;
	mov	ah,9h		;
	int	21h		;

	mov	ax,4c01h	; Abort.
	int	21h		;

#errmsg	db	'Fatal error: Bad call to real mode server.',0dh,0ah,'$'

_protret endp


;
;	_reenter - Re-enter point for real mode
;
;	This routine is called from protected mode.  Its job is
;	to restore the real mode segment registers, stack, and PSP
;	and then to call the server.
;

	public	_reenter

_reenter proc	far

;	First load the segment:offset address of the communcations buffer
;	into EDX.  We do this now because the address is on the stack and
;	we are about to switch stacks.

	mov	bp,sp		; Load from the stack, the far pointer to 
	mov	edx,4[bp]	;    communications buffer into EDX.

;	Now switch stacks and restore the real mode segment registers.

	mov	spsave2,sp	; Save the stack pointer (SS:SP) given to us 
	mov	sssave2,ss	;    by 386|DOS-Extender.

	mov	ds,dssave	; Restore the real mode registers and switch
	mov	es,essave	;    stacks.
	mov	ss,sssave	;
	mov	sp,spsave	;
	mov	bp,bpsave	;

;	Switch to the real mode PSP so that MS-DOS is all happy.

	mov	ah,50h		; Restore the real mode PSP.
	mov	bx,realpsp	;
	int	21h		;

;	Call the server with the address of the commuications buffer
;	pushed onto the stack as an argument

	push	edx		; Call the graphics server.
	call	_gserver	;

;	Switch back to the protected mode PSP so that MS-DOS stays happy

	mov	ah,50h		; Switch back to the protected mode PSP.
	mov	bx,protpsp	;
	int	21h		;

;	Switch back to the stack provided by 386|DOS-Extender and then
;	return back to the caller that is running in protected mode.

	mov	ss,sssave2	; Switch back to the stack provided by
	mov	sp,spsave2	;    386|DOS-Extender.

	ret			; Return back to protected mode caller.

_reenter endp

codeseg	ends

	end
